package Reaper::Stash;

use Moose;
use IO::File;
use Carp;
use Log::Log4perl qw(get_logger);
use DateTime;
use DateTime::TimeZone::Tzfile;
use YAML;
use Reaper::Node;

has 'nodemap'       => (is => 'ro', isa => 'HashRef', default => sub { {} });
has 'next'          => (is => 'rw', isa => 'Str');
has 'kill'          => (is => 'rw', isa => 'ArrayRef', default => sub { [] });
has 'now'           => (is => 'ro', isa => 'DateTime');
has 'tz'            => (is => 'ro', default => sub {
  DateTime::TimeZone::Tzfile->new('/etc/localtime');
});
has 'noop'          => (is => 'ro', isa => 'Bool');

sub kill_list 
{
  my $self = shift;
  map { $self->nodemap->{$_} } @{$self->kill};
}

sub node_list 
{
  my $self = shift;
  values %{$self->nodemap};
}

# Kill all the nodes in the kill list
sub killall
{
  my $self = shift;
  my $log = get_logger('');
  my @kill = $self->kill_list;
  $log->debug("stash: found " . scalar(@kill) . ' job(s) to kill');
  return unless @kill;
  my @unkilled = ();
  foreach (@kill) {
    my $rv = $_->kill();
    $log->debug(sprintf 'kill on %s returned %s %s', 
      $_->file, 
      $rv ? 'success' : 'failure', 
      $rv ? '' : " ($rv)");
    push @unkilled, $_ if ! $rv && $rv eq 'EKILLFAILED';
  }
  $self->kill( \@unkilled );
}

sub nextnode
{
  my $self = shift;
  return $self->next ? $self->nodemap->{ $self->next } : '';
}

# Update kill and next
sub update 
{
  my $self = shift;
  my $log = get_logger('');

  $self->kill([]);
  $self->next('');

  my $now = $self->now || DateTime->now( time_zone => $self->tz );
  my $now_time = $now->strftime('%H:%M');
  my %today = ( year => $now->year, month => $now->month, day => $now->day, time_zone => $self->tz );
  for my $file (keys %{$self->nodemap}) {
    my $time = $self->nodemap->{$file}->time();
    my ($hh, $mi) = split /:/, $time;
    my $dt = DateTime->new( %today, hour => $hh, minute => $mi );
    if ($dt <= $now) {
      $log->debug("$file time before now ($time <= $now_time) - adding to kill list");
      push @{$self->{'kill'}}, $file;
    }
    else {
#     $log->debug("$file time after now ($time > $now_time) - checking for next");
      if (my $next = $self->nextnode) {
        my $next_dt = $next->date;
        if ($dt < $next_dt) {
          $log->debug("$file time ($time) precedes next ($next_dt) - replacing");
          $self->next($file);
        }
        else {
          $log->debug("$file time ($time) after next ($next_dt) - ignoring");
        }
      }
      else {
        $log->debug("no current next - initialising with $file ($time)");
        $self->next($file);
      }
    }
  }
}

sub add_file
{
  my $self = shift;
  my ($file, %arg) = @_;
  my $log = get_logger('');

  croak "Missing file '$file'" unless -f $file;
  sleep 3 unless -s $file;
  $log->debug("stash: adding file $file");
  my $fh = IO::File->new( $file, 'r' )
    or croak "Cannot open file '$file': $!";
  my $data = <$fh>;
  $fh->close;

  # Parse data
  if (defined $data && $data =~ m/^\s*(\d{1,2}:\d{2})\s+(\d+)(?:\s+(.*))?$/) {
    my ($time, $pid, $command) = ($1, $2, $3);
    if (my $node = eval { Reaper::Node->new( 
        file => $file, 'time' => $time, pid => $pid, command => $command, tz => $self->tz,
        noop => $self->noop,
        %arg
        )}) {
      $self->nodemap->{$file} = $node;
      $log->debug("stash: time $time, pid $pid, command $command");
    }
    else {
      $log->error("invalid pid $pid - removing file $file");
      unlink $file unless $arg{fake} || $arg{noop};
    }
  }
  else {
    croak "Bad data in file '$file'";
  }
}

sub remove_file
{
  my $self = shift;
  my ($file) = @_;
  if ($self->nodemap->{$file}) {
    my $log = get_logger('');
    $log->debug("stash: removing file $file");
    delete $self->nodemap->{$file} unless $self->noop;
  }
}

sub dump 
{
  my $self = shift;
  Dump { 
    'now'       => $self->now ? $self->now->strftime('%H:%M') : '',
    'nodemap'   => { map { $_ => $self->nodemap->{$_}->short } keys %{$self->nodemap} },
    'kill'      => $self->kill(),
    'next'      => $self->nextnode ?
                   join(', ', $self->nextnode->time, 
                              $self->nextnode->time_seconds($self->now), 
                              $self->nextnode->file) : 
                   '',
  };
}

1;

__END__

=head1 NAME

Reaper::Stash - intelligent container for reaper nodes

=head1 SYNOPSIS

    $stash = Reaper::Stash->new;
    $stash->add_file($_) foreach glob "$watch_dir/*";
    $stash->update;
    $stash->killall;
    print $stash->next;
    $next = $stash->nextnode;

=head1 AUTHOR

Gavin Carr <gavin@openfusion.com.au>

=head1 COPYRIGHT

Copyright 2008 Gavin Carr.

This is free software licensed under the terms of the Perl Artistic Licence.

=cut

